﻿:Class PubSub
⍝ WSServer plugin that implements a publish/subscribe interface

    ⎕ML←⎕IO←1

    ∇ r←Version
      :Access public shared
      r←'PubSub' '0.5.0' '2023-10-01'
    ∇

    :field public Config        ⍝ configuration namespace
    :field public CodeLocation  ⍝ Location

    :field public AuthenticateFn←''

    :field public Channels      ⍝ vector of channel names
    :field public Actions       ⍝ vector of action names
    :field public Server        ⍝ reference to WSServer instance

    :field public Connections   ⍝ reference to _connections in WSServer
    :field public Subscriptions ⍝ one per user

⍝ PubSub protocol:
⍝ Actions: Login, Logout, Subscribe, Unsubscribe
⍝ action: 'Subscribe', channel: 'channel_name'
⍝ action: 'Unsubscribe', channel: 'channel_name'
⍝ action: 'ListChannels'


    ∇ (rc msg)←New arg
    ⍝ create a new instance of Subscribe
      :Access public shared
      msg←''
      :Trap 777
          :If 0∊⍴arg
              rc←##.⎕NEW ⎕THIS
          :Else
              rc←##.⎕NEW ⎕THIS arg
          :EndIf
      :Else
          (rc msg)←¯1 ⎕DMX.EM
      :EndTrap
    ∇

    ∇ make
      :Access public
      :Implements constructor
      MakeCommon
    ∇

    ∇ make1 args;mask;channels;actions
      :Access public
      :Implements constructor
      MakeCommon
      'PubSub constructor argument is not a namespace'⎕SIGNAL 777/⍨9.1≠nameClass args
      'No PubSub channels defined'⎕SIGNAL 777/⍨0=args.⎕NC'Channels'
      :If ∨/mask←3≠⌊CodeLocation.⎕NC channels←args.Channels.Name
          ('PubSub channel function',(1=+/mask)↓'s missing:',∊⍕mask/channels)⎕SIGNAL 777
      :Else
          Channels←⎕JSON⍣2⊢args.Channels
      :EndIf
      :If 0=args.⎕NC'Actions'
      :OrIf 0∊⍴args⍎'Actions'
          Actions←0⍴⊂''
      :Else
          :If ∨/mask←3≠⌊CodeLocation.⎕NC actions←args.Actions.Name
              ('PubSub action function',(1=+/mask)↓'s missing:',∊⍕mask/actions)⎕SIGNAL 777
          :Else
              Actions←⎕JSON⍣2⊢args.Actions
          :EndIf
      :EndIf
      :If 0≠args.⎕NC'AuthenticateFn'
          AuthenticateFn←args.AuthenticateFn
      :EndIf
      Config←args
    ∇

    ∇ MakeCommon
    ⍝ sets Server to the creating WebSocketServer instance, or # (for debugging)
      Server←{9.2=nameClass ⍵:⍵ ⋄ #}##  ⍝ backpointer to server (or root if not running through server)
      CodeLocation←{6::⍵ ⋄ ⍵⍎'CodeLocation'}Server
      :Trap 11
          JSONin←0 ##.##.⎕JSON⍠('Dialect' 'JSON5')⊢ ⋄ {}JSONin'1'
          JSONout←1 ##.##.⎕JSON⍠'HighRank' 'Split'⊢ ⋄ {}JSONout 1
      :Else ⍝ fallback if using a Dyalog version that doesn't support HighRank or Dialect
          JSONin←0 ##.##.⎕JSON⍠⊢
          JSONout←1 ##.##.⎕JSON⊢
      :EndTrap
     
    ∇

    ∇ rc←obj onWSReceive payload;action
    ⍝ obj is Conga object name
    ⍝ payload is the message payload
      :Access public
      rc←¯1
      :Trap 11 ⍝ domain error
          payload←JSONin payload
      :Else
          →0
      :EndTrap
      →0⍴⍨9.1≠nameClass payload ⍝ ensure it's a namespace
      →0⍴⍨0=payload.⎕NC'Action' ⍝ must have an Action
      →0⍴⍨0∊⍴payload.Action     ⍝ can't be empty
      :If payload.Action≡AuthenticateFn
          rc←obj Authenticate payload
      :Else
          →0⍴⍨~(⊂payload.Action)∊Actions ⍝ must be in the list of Actions
          rc←obj(CodeLocation⍎payload.Action)payload  ⍝ execute the action
      :EndIf
    ∇

    ∇ {rc}←wsid Authenticate payload
      rc←wsid(CodeLocation⍎AuthenticateFn)payload
    ∇

    ∇ {rc}←wsid WSSend msg
      :Access public
      rc←wsid ##.WSSend msg
    ∇

    ∇ {rc}←wsid SetUserID userid;ind
      :Access public
      rc←¯1
      :Hold '_connections'
          →0⍴⍨0=ind←Connections.index[1;]{i×(≢⍺)≤i←⍺⍳⍵}⊂wsid
          Connections.index[3;ind]←⊂userid
          rc←0
      :EndHold
    ∇

    ∇ r←isJSON w
      :Access public
      r←0
      →0⍴⍨(⎕DR'')≠⎕DR w ⍝ not character? can't be JSON
      →0⍴⍨~(⊃w)∊'-{["',⎕D ⍝ doesn't begin with '-{["'⎕D? can't be JSON
      r←{0::0 ⋄ 1⊣0 JSONin ⍵}w ⍝ test for JSONableness fails on APL that looks like JSON (e.g. '"abc"')
    ∇

    nameClass←{⎕NC⊂,'⍵'}


    :Section PubSub API

    ∇ r←CurrentUsers
      :Access public
      r←⍉Connections.index[1 3;]
    ∇

    ∇ wsid←WSids id
    ⍝ return WebSocket ids
    ⍝ if id is a user name, multiple wsids might be returned
      :Access public
      :If 0∊⍴'.CON\d{8}$'⎕S{⍵}⊢id
          id←(Connections.index[3;]≡¨⊆,id)/Connections.index[1;]
      :EndIf
      wsid←,⊆id
    ∇

    ∇ rc←{ids}Publish(channel payload);msg;wsid;wsids
      :Access public
      msg←⎕NS''
      msg.Action←channel
      'msg'⎕NS toAPL payload
      :If 0=⎕NC'ids'
      :OrIf 0∊⍴ids
          wsids←Connections.index[1;]
      :Else
          wsids←WSids ids
      :EndIf
      :For wsid :In wsids
          wsid WSSend msg
      :EndFor
    ∇

    ∇ apl←toAPL r
      :Access public
      apl←r
      :If isJSON apl
          apl←JSONin apl
      :EndIf
      :If 9.1≠nameClass apl
          (apl←⎕NS'').Payload←apl
      :EndIf
    ∇

    ∇ userid←UserIDforConnection conx
      :Access public
      userid←(Connections.index[1;]≡¨⊆,conx)/Connections.index[3;]
    ∇
    :endsection
:EndClass
